#include "main.hpp"

volatile uint8_t time_per_round_pointer = 0;
volatile double time_per_round_ms[TIME_PER_ROUND_VALS];

time_t run_time = 0;
time_t start_time = 0;
hw_timer_t * timer = NULL;
int ser_esc_output = -1;
double last_error = 0.0;

void ISR_HALL1();
void ISR_HALL2();
void ISR_HALL3();
void ISR_HALL4();
void ISR_HALL5();
void ISR_HALL6();
void ISR_HALL7();
void ISR_HALL8();
void ISR_onTimer();

void data_generate();
void data_store();
void data_check();
void data_init();
void speed_set();
void eval_max_min_speed();

void PID();
void limit_values(double calc_value, double *pvalue, double factor,  double min_limit , double max_limit);
unsigned long  time_per_round_calc();
void count_secs(time_t* run_time);

void get_serial_cmd();

#ifdef HAS_MQTT
void json_init();
#endif

void setup() 
{

  delay(1000);
  Serial.begin(115200);
  Serial.println("Init capture portal ...");

  randomSeed(micros());

  display.begin();
  display.progressBar(0,"Lets start ...", true);

  capportal.begin();
  html_content = "<!DOCTYPE html><html><head><meta http-equiv=\"refresh\" content=\"0; url='/_ac'\" /></head><body></body></html>";  

  #ifdef HAS_MQTT
  mqtt.begin(mqttServer, mqttPort, mqttUser, mqttPassword);
  #endif

  Store.begin();
  
  Serial.println("Init ETC ...");
  ESC.attach(ESC_PIN,1000,2000); 
  display.progressBar(0,"Init controller", true);
  ESC.write(20);
  delay(2000);
  display.progressBar(1,"Connect ETC!", "Please connnect the ETC", "controller to power!", true);  
  delay(5000);
  display.progressBar(2,"Calibrating ETC", "You may hear some beeps.", "If not - check ETC connection.", true);    
  ESC.write(0);
  for(uint8_t i=3;i<=100;i=i+10)
  {
    display.progressBar(i,false);
    delay(1000);
  }
  ESC.write(0);
  display.progressBar(100,"ETC init done!", true);
  delay(1000);  
  display.progressBar(100,"Init data and IRQ!", true);
  data_init();

  #ifndef DUMMY_DATA

  pinMode(39, INPUT_PULLDOWN);
  pinMode(HALL1_PIN, INPUT);
  pinMode(HALL2_PIN, INPUT);
  pinMode(HALL3_PIN, INPUT);
  pinMode(HALL4_PIN, INPUT);
  pinMode(HALL5_PIN, INPUT);
  pinMode(HALL6_PIN, INPUT);
  pinMode(HALL7_PIN, INPUT);
  pinMode(HALL8_PIN, INPUT);

  Serial.println("Init hall sensor interrupts ...");
  attachInterrupt(digitalPinToInterrupt(HALL1_PIN), ISR_HALL1, FALLING);
  attachInterrupt(digitalPinToInterrupt(HALL2_PIN), ISR_HALL2, FALLING);
  attachInterrupt(digitalPinToInterrupt(HALL3_PIN), ISR_HALL3, FALLING);
  attachInterrupt(digitalPinToInterrupt(HALL4_PIN), ISR_HALL4, FALLING);
  attachInterrupt(digitalPinToInterrupt(HALL5_PIN), ISR_HALL5, FALLING);
  attachInterrupt(digitalPinToInterrupt(HALL6_PIN), ISR_HALL6, FALLING);
  attachInterrupt(digitalPinToInterrupt(HALL7_PIN), ISR_HALL7, FALLING);
  attachInterrupt(digitalPinToInterrupt(HALL8_PIN), ISR_HALL8, FALLING);  

  #endif

  Serial.println(" Measuring min and max speed values ... ");
  eval_max_min_speed();

  #ifdef HAS_MQTT
    display.progressBar(100,"Init done!", "Connection to MQTT ...", true);
  #else
    display.progressBar(100,"Init done!", "Starting ...", true);
  #endif
  Serial.println(" Auf gehts ... ");
}

void loop() 
{
  capportal.handle();
  get_serial_cmd();
  data_generate();
  count_secs(&run_time);
  speed_set();
  data_check();
  data_store();

  display.show_values(Pid_params.speed, Pid_params.min_speed, Pid_params.max_speed, HallData, HALL_SENSORS_COUNT, run_time, capportal.localIP());
  
  #ifdef HAS_MQTT
  mqtt.send_and_loop("infinity/data", json.create_json_string(AllData, ALL_DATA_COUNT, &Pid_params, run_time));
  #endif
 
}

void eval_max_min_speed()
{
  display.progressBar(0,"Min/Max measure!", true);
  Pid_params.min_esc_out = 0;
  Pid_params.min_speed = 0;
  for(int out = Pid_params.min_speed; out < Pid_params.max_esc_out; out++)
  {
    ESC.write(out);
    String esc_str = String("esc_out = " + String(out));
    String speed_str = String("speed_out = " + String(HallData[0].value));
    delay(1000);
    for(uint8_t i=0;i<10;i++)
    {
      display.progressBar(out*10+i,"Min/Max measure!", esc_str.c_str(), speed_str.c_str(), true);
      data_check();
      delay(100);

      if (0 == Pid_params.min_speed && HallData[0].value > 0)
      {
        Pid_params.min_speed = HallData[0].value;
        Pid_params.min_esc_out = out;
        break;
      }
    }
    if (Pid_params.min_speed > 0)
    {
      break;
    }
  }
  display.progressBar(100,"Min speed/esc:", String("min_speed =" + String(Pid_params.min_speed) ).c_str(), String("min_esc_out =  " + String(Pid_params.min_esc_out)).c_str(), true);
  ESC.write(Pid_params.max_esc_out);
  delay(2000);
  data_check();
  Pid_params.max_speed = HallData[0].value;
  display.progressBar(100,"Max speed/esc:", String("max_speed =" + String(Pid_params.max_speed) ).c_str(), String("max_esc_out =  " + String(Pid_params.max_esc_out)).c_str(), true);
  ESC.write(0);
  Serial.printf(">min_speed:%d\n>max_speed:%d\n>min_esc_out:%f\n>max_esc_out:%f\n",
                 Pid_params.min_speed, 
                 Pid_params.max_speed, 
                 Pid_params.min_esc_out, 
                 Pid_params.max_esc_out );
  delay(2000);
}

void count_secs(time_t* run_time)
{
  static time_t last_update_ms = 0;
  if(last_update_ms == 0)
  {
    last_update_ms = millis();
  }
  if(millis() - last_update_ms >= 1000)
  {
    last_update_ms = millis();
    *run_time = *run_time + 1;
  }
}

void limit_values(double calc_value, double *pvalue, double factor,  double min_limit , double max_limit)
{
  if(factor != 0)
  {
    if(factor*calc_value > max_limit)
    {
      *pvalue = max_limit/factor;
    } 
    else if(factor*calc_value < min_limit)
    {
      *pvalue = min_limit/factor;
    }
    else 
    {
      *pvalue = calc_value;
    }
  }
  else
  {
    *pvalue = 0;
  }
}

void PID()
{
  static unsigned long last_pid_ms = millis();

  Pid_params.current_speed = HallData[0].value;
  Pid_params.err = constrain(Pid_params.speed - Pid_params.current_speed, -1*Pid_params.max_speed, Pid_params.max_speed);
  limit_values(Pid_params.integ + Pid_params.err, &Pid_params.integ, Pid_params.ki,  MIN_I_VALUE , MAX_I_VALUE);
  limit_values(Pid_params.err - last_error, &Pid_params.derivative, Pid_params.kd,  MIN_D_VALUE , MAX_D_VALUE);  
  last_error = Pid_params.err;

  Pid_params.output = Pid_params.kp * Pid_params.err + Pid_params.ki * Pid_params.integ + Pid_params.kd * Pid_params.derivative;
  if(-1 == ser_esc_output ) 
  {
    if(Pid_params.start_done == true)
    {
      last_pid_ms = millis();
      Pid_params.esc_output = int(Pid_params.min_esc_out + constrain(Pid_params.output, 0, Pid_params.max_esc_out - Pid_params.min_esc_out));
    }
    else
    {
      double rise = (double(Pid_params.max_esc_out - Pid_params.min_esc_out)/ double(Pid_params.max_speed - Pid_params.min_speed));
      double offset = double(Pid_params.max_esc_out) - rise*Pid_params.max_speed;
      Pid_params.esc_output = int(rise*Pid_params.speed + offset);
      Pid_params.derivative = 0;
      Pid_params.integ = 0;
      Serial.printf("Start speed ... rise=%6.3f offset=%6.3f esc_out=%d\n", rise, offset, Pid_params.esc_output);
      if((millis() - last_pid_ms > 2000) || (abs(Pid_params.current_speed - Pid_params.speed)>= (0.1*Pid_params.speed)) )
      {
        Pid_params.start_done = true;
      }
    }
  }
  else
  {
    Pid_params.esc_output = ser_esc_output;
  }

}

unsigned long time_per_round_calc()
{
  unsigned long mid_time = 0;
  unsigned long min_time = 0;
  unsigned long max_time = 0;
  uint8_t nr_times = 0;
  bool real_values = false;
  if(millis() - HallData[0].act_update_ms <= (HALL_NR_TURN*MAX_TIME_PER_TURN_MS))
  {
    real_values = true;
    for(uint8_t i = 0; i < TIME_PER_ROUND_VALS; i++)
    {
      unsigned long time_round_ms = time_per_round_ms[i]/HALL_NR_TURN;
      if(time_round_ms > 0)
      {
        if(time_round_ms < min_time || min_time == 0)
        {
          min_time = time_round_ms;
        }
        if(time_round_ms > max_time)
        {
          max_time = time_round_ms;
        }
        nr_times++;
        mid_time = mid_time + time_round_ms;
      }
    }
    if(nr_times >= 3)
    {
      mid_time = (mid_time - max_time  - min_time)/(nr_times - 2);
    }
    else if (nr_times > 0)
    {
      mid_time = mid_time/nr_times;
    }
  }
  else
  {
    for (uint8_t i= 0; i < TIME_PER_ROUND_VALS; i++)
    {
      time_per_round_ms[i] = HALL_NR_TURN*MAX_TIME_PER_TURN_MS;
    }     
    mid_time = MAX_TIME_PER_TURN_MS;
  }
  /*
  static unsigned long last_mid_time = millis();
  if(millis() - last_mid_time > 500)
  {
    Serial.printf(">turn_time_ms:%d\n>real_values:%d\n", mid_time, real_values);
    last_mid_time = millis();
  }
  */
  return(mid_time);
}

void speed_set()
{
  static unsigned long last_esc_write_ms = 0;
  if(0 == last_esc_write_ms || millis() - last_esc_write_ms > UPDATE_ESC_EVERY_MS)
  {
    PID();
    Serial.printf(">speed:%06.3f\n>current:%06.3f\n>esc:%06d\n>prop:%06.3f\n>derive:%06.3f\n>integ:%06.3f\n",
                  Pid_params.speed, 
                  Pid_params.current_speed, 
                  Pid_params.esc_output, 
                  Pid_params.err * Pid_params.kp, 
                  Pid_params.derivative* Pid_params.kd, 
                  Pid_params.integ* Pid_params.ki);
    //Serial.printf(">speed:%06.3f\n>current:%06.3f\n>esc:%06.3f\n",Pid_params.speed, Pid_params.current_speed, Pid_params.esc_output);
    ESC.write(Pid_params.esc_output);
    last_esc_write_ms = millis();
    Pid_params.timestamp = last_esc_write_ms;      
  }
}

void data_init()
{
  Serial.println("Init internal data ...");
  //kp=0.006000 ki=0.002000 kd=0.006000
  Pid_params.timestamp = millis();
  Pid_params.integ = 0.0;
  Pid_params.derivative = 0.0;
  Pid_params.esc_output = 0;
  Pid_params.speed = 1000.0;
  Pid_params.current_speed = 0.0;
  Pid_params.err = 0;
  Pid_params.output = 0.0;
  Pid_params.kp = 0.006;
  Pid_params.ki = 0.002;
  Pid_params.kd = 0.006;
  Pid_params.max_speed = 5000;
  Pid_params.min_speed = 0;
  Pid_params.max_esc_out = MAX_ESC_OUT;
  Pid_params.min_esc_out = MIN_ESC_OUT;
  Pid_params.start_done = false;
  #ifdef MACHINE_RESTART
    run_time = 0;
    start_time = capportal.epochTime();
    Store.set_int(EEPROM_NAME_RUN_TIME , run_time );
    Store.set_int(EEPROM_NAME_START_EPOCHTIME, start_time );
    Serial.printf("Machine restarted. Set runtime to 0. Stored this value and the epoch start time %lu to EEPROM.\n", start_time);
  #else
    run_time = Store.get_int(EEPROM_NAME_RUN_TIME );
    start_time = Store.get_int(EEPROM_NAME_START_EPOCHTIME );
    Serial.printf("Machine runtime loaded from EPPROM: %lu seconds. Machine was started %s\n", run_time, capportal.localTime(start_time));
  #endif
  for (uint8_t i= 0; i < TIME_PER_ROUND_VALS; i++)
  {
    time_per_round_ms[i] = HALL_NR_TURN*MAX_TIME_PER_TURN_MS;
  } 
  unsigned long current_millis = millis();
  for (uint8_t i = 0; i < ALL_DATA_COUNT; i ++)
  {
    AllData[i].name = Data_names[i];
    AllData[i].unit = Data_units[i];
    AllData[i].value = 0;
    AllData[i].unit_factor = Data_units_factor[i];
    AllData[i].period_ticks = 0;
    #ifdef MACHINE_RESTART
    AllData[i].ticks = 0;
    AllData[i].timestamp = 0;
    #else
    AllData[i].ticks = Store.get_int(EEPROM_NAME_TICK+String(i));
    AllData[i].timestamp = Store.get_int(EEPROM_NAME_TIMESTAMP+String(i));
    if(i != 0)
    {
      AllData[i].value = (AllData[i].ticks * AllData[i].unit_factor);
    }
    #endif
    AllData[i].act_update_ms = 0;
  }
}


void data_check()
{
  /*
  Serial.printf("H1(pin %d): %d, H2(pin %d): %d, H3(pin %d): %d, H4(pin %d): %d, H5(pin %d) : %d, H6(pin %d) : %d, H7(pin %d) : %d, H8(pin %d) : %d.\n",
                HALL1_PIN, digitalRead(HALL1_PIN),
                HALL2_PIN, digitalRead(HALL2_PIN),
                HALL3_PIN, digitalRead(HALL3_PIN),
                HALL4_PIN, digitalRead(HALL4_PIN),
                HALL5_PIN, digitalRead(HALL5_PIN),
                HALL6_PIN, digitalRead(HALL6_PIN),
                HALL7_PIN, digitalRead(HALL7_PIN),
                HALL8_PIN, digitalRead(HALL8_PIN)                
                );
  
  */

  for(uint8_t nr = 0; nr < HALL_SENSORS_COUNT; nr++)
  {
    if( 0 == nr)
    {
      unsigned long mid_time = time_per_round_calc();
      if( mid_time != 0)
      {
        if( mid_time >= MAX_TIME_PER_TURN_MS)
        {
          HallData[nr].value = 0;
        }
        else
        {
          HallData[nr].value = HallData[nr].unit_factor / mid_time;
          HallData[nr].value = HallData[nr].value > Pid_params.max_speed? Pid_params.max_speed:HallData[nr].value;
        }
      }
    }
    else{
      HallData[nr].value = (HallData[nr].ticks * HallData[nr].unit_factor);
    }
  }
}

void data_store()
{
  static unsigned long _last_stored_ms = 0;
  if(_last_stored_ms == 0)
  {
    _last_stored_ms = millis();
  }
  if(millis() - _last_stored_ms >= EEPROM_STORE_RUNTIME_ALL_MS)
  {
    _last_stored_ms = millis();
    Store.set_int(EEPROM_NAME_RUN_TIME, run_time);
    for(uint8_t nr = 0; nr < HALL_SENSORS_COUNT; nr++)
    {
      Store.set_int(EEPROM_NAME_TICK+String(nr), HallData[nr].ticks);
      Store.set_int(EEPROM_NAME_TIMESTAMP+String(nr), HallData[nr].timestamp);
    }
  }
}


void get_serial_cmd()
{
  if(Serial.available()){
        String command = Serial.readStringUntil('\n');
        Serial.printf("Received command '%s'", command.c_str());
        if(command.startsWith("kp=")){
            Pid_params.kp = command.substring(command.indexOf("=")+1).toDouble();
            Serial.printf("Set kp to %f\n", Pid_params.kp);
        }
        else if(command.startsWith("ki=")){
            Pid_params.integ = 0.0;
            Pid_params.ki = command.substring(command.indexOf("=")+1).toDouble();
            Serial.printf("Set ki to %f\n", Pid_params.ki);
        }
        else if(command.startsWith("speed=")){
            Pid_params.start_done = false;
            Pid_params.speed = command.substring(command.indexOf("=")+1).toDouble();
            Serial.printf("Set speed to %f\n", Pid_params.speed);
        }
        else if(command.startsWith("kd=")){
            Pid_params.derivative = 0.0;
            Pid_params.kd = command.substring(command.indexOf("=")+1).toDouble();
            Serial.printf("Set kd to %f\n", Pid_params.kd);
        }
        else if(command.startsWith("esc=")){
            ser_esc_output = command.substring(command.indexOf("=")+1).toDouble();
            Serial.printf("Set esc_output to %f\n", ser_esc_output);
        }

        else if(command.equals("params")){
            Serial.printf("Actual values: kp=%f  ki=%f  kd=%f  \n", Pid_params.kp, Pid_params.ki, Pid_params.kd);
        }
        else if(command.equals("control")){
            ser_esc_output = -1;
            Serial.printf("Set to drive PID control mode.\n");
        }
        else if(command.equals("max")){
            ser_esc_output = Pid_params.max_esc_out;
            Serial.printf("Set to maximal speed. %f\n", ser_esc_output);
        }        
        else if(command.equals("min")){
            ser_esc_output = Pid_params.min_esc_out;
            Serial.printf("Set to minimal speed %f .\n", ser_esc_output );
        }
        else if(command.equals("start")){
            ser_esc_output = -1;
            Serial.printf("Start drive PID control mode.\n");
        }
        else if(command.equals("stop")){
            ser_esc_output = 0;
            Serial.printf("Stop drive.\n");
        }
        else if(command.equals("pidreset")){
            Pid_params.err = 0.0;
            last_error = 0.0;
            Pid_params.integ = 0.0;
            Pid_params.derivative = 0.0;
            Serial.printf("Reset PID variables.\n");
        }

        else if(command.equals("help")){
            Serial.printf("Commands: kp,ki,kd=x.xxxx | params | control | max | min | start | stop | speed=x | esc=x | reset | help \n");
        }
        else{
            Serial.printf("Invalid command.");
        }
  }
}

#ifndef DUMMY_DATA
void IRAM_ATTR ISR_HALL1()
{
  const uint8_t hallnr = 0;
  const unsigned long hall_ticks =  HALL_NR_TURN * HALL_TICKS_PER_TURN;
  const unsigned long hall_time =  HALL_NR_TURN * MAX_TIME_PER_TURN_MS;
  HallData[hallnr].ticks++;
  HallData[hallnr].period_ticks++;
  HallData[hallnr].timestamp = run_time;
  // each 4 ticks is one turn
  // so we calculate speed each 4 ticks... just in case we check for bigger ... 
  if(HallData[hallnr].period_ticks >= hall_ticks ){
    unsigned long time_ms = millis();
    unsigned long time_diff_ms = time_ms - HallData[hallnr].act_update_ms;
    time_diff_ms= time_diff_ms > hall_time?hall_time:time_diff_ms;
    HallData[hallnr].period_ticks = 0;
    time_per_round_ms[time_per_round_pointer] = time_diff_ms;
    HallData[hallnr].act_update_ms = time_ms;
    time_per_round_pointer++;
    time_per_round_pointer = time_per_round_pointer>TIME_PER_ROUND_VALS?0:time_per_round_pointer;
  }
}

void IRAM_ATTR ISR_HALL2()
{
  const uint8_t hallnr = 1;
  HallData[hallnr].ticks++;
  HallData[hallnr].timestamp = run_time;
  HallData[hallnr].act_update_ms = millis();
}

void IRAM_ATTR ISR_HALL3()
{
  const uint8_t hallnr = 2;
  HallData[hallnr].ticks++;
  HallData[hallnr].timestamp = run_time;
  HallData[hallnr].act_update_ms = millis();
}

void IRAM_ATTR ISR_HALL4()
{
  const uint8_t hallnr = 3;
  HallData[hallnr].ticks++;
  HallData[hallnr].timestamp = run_time;
  HallData[hallnr].act_update_ms = millis();
}

void IRAM_ATTR ISR_HALL5()
{
  const uint8_t hallnr = 4;
  HallData[hallnr].ticks++;
  HallData[hallnr].timestamp = run_time;
  HallData[hallnr].act_update_ms = millis();
}

void IRAM_ATTR ISR_HALL6()
{
  const uint8_t hallnr = 5;
  HallData[hallnr].ticks++;
  HallData[hallnr].timestamp = run_time;
  HallData[hallnr].act_update_ms = millis();
}

void IRAM_ATTR ISR_HALL7()
{
  const uint8_t hallnr = 6;
  HallData[hallnr].ticks++;
  HallData[hallnr].timestamp = run_time;
  HallData[hallnr].act_update_ms = millis();
}

void IRAM_ATTR ISR_HALL8()
{
  const uint8_t hallnr = 7;
  HallData[hallnr].ticks++;
  HallData[hallnr].timestamp = run_time;
  HallData[hallnr].act_update_ms = millis();
}
#endif

//this code is just for generating dummy data - in normal use not needed
void data_generate()
{
  #ifdef DUMMY_DATA

  static unsigned long _last_updated_ms = 0; 

  speed = map(esc_output, MIN_ESC_SPEED, MAX_ESC_SPEED, MIN_SPEED, MAX_SPEED) + random(-50,50);

  unsigned long _wait_ms = map( speed, MIN_SPEED, MAX_SPEED, 60000/MIN_SPEED/4, 60000/MAX_SPEED/4);
  //Serial.printf("Speed is : %lu and _wait_ms is %lu\n", speed, _wait_ms);

  if(_last_updated_ms == 0)
  {
    _last_updated_ms = millis();
  }

  if(millis() - _last_updated_ms >= _wait_ms )
  {
    _last_updated_ms = millis();

    for(uint8_t i = 0; i < HALL_SENSORS_COUNT; i++)
    {
      if(i == 0)
      {
        HallData[i].ticks++;
        HallData[i].period_ticks++;
        HallData[i].timestamp = run_time;

        // each 4 ticks is one turn
        // so we calculate speed each 4 ticks... just in case we check for bigger ... 
        if(HallData[i].period_ticks >= (HALL_NR_TURN * HALL_TICKS_PER_TURN))
        {
          unsigned long time_ms = millis();
          HallData[i].period_ticks = 0;
          time_per_round_ms[time_per_round_pointer] = time_ms - HallData[i].act_update_ms;
          Serial.printf(">tick_time:%d\n", time_per_round_ms[time_per_round_pointer]);
          HallData[i].act_update_ms = time_ms;
          time_per_round_pointer++;
          if(time_per_round_pointer > TIME_PER_ROUND_VALS)
          {
            time_per_round_pointer = 0;
          }
        }
      }
      else
      {
        //Serial.printf("i=%d HallData[i-1].ticks=%lu %4 = %d\n" , i, HallData[i-1].ticks , HallData[i-1].ticks % 4);
        if (HallData[i].act_update_ms != HallData[i-1].act_update_ms && HallData[i-1].ticks % 4 == 0)
        {
          HallData[i].ticks++;
          HallData[i].act_update_ms = HallData[i-1].act_update_ms;
          HallData[i].timestamp = run_time;
        }
      }
    }
  }
  #endif
}
